/* icmplog.c of iplog v1.4 -- eric@ojnk.org */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <string.h>
#include <netdb.h>
#define __FAVOR_BSD
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "include/icmp-compat.h"
#include "include/iplog.h"
#include "include/lockfile.h"

static char *protolookup(int);
static char *print_time(u_int);
static char *my_ntoa(struct in_addr);

struct packet {
	struct ip ip;
	struct icmp icmp;
	struct tcphdr tcp;
};

#ifdef DETECT_SMURF
#include "include/smurf.h"
#endif

#define ICMP_SIZE	sizeof(struct ip)

#define proto	pkt.icmp.icmp_ip.ip_p
#define gw		pkt.icmp.icmp_gwaddr
#define src		pkt.ip.ip_src
#define dest	pkt.icmp.icmp_ip.ip_dst
#define orig	pkt.icmp.icmp_ip.ip_src

static const char *codes[] =
{
	"echo reply from %s sz=%u(+%u)",	/* ICMP_ECHOREPLY */
	"undefined icmp type 1 recieved from %s(?)",	/* ? */
	"undefined icmp type 2 recieved from %s(?)",	/* ? */
	"destination unreachable from %s",	/* ICMP_DEST_UNREACH */
	"source quench from %s",	/* ICMP_SOURCE_QUENCH */
	"redirect message from %s",	/* ICMP_REDIRECT */
	"undefined icmp type 6 recieved from %s(?)",	/* ? */
	"undefined icmp type 7 recieved from %s(?)",	/* ? */
	"echo from %s sz=%u(+%u)",	/* ICMP_ECHO */
	"router advertisement from %s",		/* ICMP_ROUTERADVERT (rfc1256) */
	"router solicitation from %s",	/* ICMP_ROUTERSOLICIT (rfc1256) */
	"time exceeded (%s)",		/* ICMP_TIME_EXCEEDED */
	"parameter problem from %s (ptr %d)",	/* ICMP_PARAMETERPROB */
	"timestamp request from %s",	/* ICMP_TIMESTAMP */
	"timestamp reply from %s (%s)",		/* ICMP_TIMESTAMPREPLY */
	"info request from %s",		/* ICMP_INFO_REQUEST */
	"info reply from %s",		/* ICMP_INFO_REPLY */
	"address mask request from %s",		/* ICMP_ADDRESS */
	"address mask reply from %s (%s)"	/* ICMP_ADDRESSREPLY */
};

static const char *unreach[] =
{
	"%s: network is unreachable [%s: dp=%u sp=%u]",		/* ICMP_NET_UNREACH */
	"%s: host is unreachable [%s: dp=%u sp=%u]",	/* ICMP_HOST_UNREACH */
	"%s: %s: protocol is unreachable",	/* ICMP_PROT_UNREACH */
	"%s: %s port is unreachable [dp=%u sp=%u]",		/* ICMP_PORT_UNREACH */
	"%s: fragmentation needed (IP_DF set)",		/* ICMP_FRAG_NEEDED */
	"%s: source route failed [%s: dp=%u sp=%u]",	/* ICMP_SR_FAILED */
	"%s: network is unreachable(?) [%s: dp=%u sp=%u]",	/* ICMP_NET_UNKNOWN */
	"%s: host is unreachable(?) [%s: dp=%u sp=%u]",		/* ICMP_HOST_UNKNOWN */
	"%s: host is isolated from source",		/* ICMP_HOST_ISOLATED */
	"%s: network is unreachable (admin prohibited) ",	/* ICMP_NET_ANO */
	"%s: host is unreachable (admin prohibited)",	/* ICMP_HOST_ANO */
	"%s: network is unreachable (tos is prohibited)",	/* ICMP_NET_UNR_TOS */
	"%s: host is unreachable (tos is prohibited)",	/* ICMP_HOST_UNR_TOS */
	"%s: packet filtered [%s: dp=%u sp=%u]",	/* ICMP_PKT_FILTERED */
	"%s: precedence violation",	/* ICMP_PREC_VIOLATION */
	"%s: precedence cutoff"		/* ICMP_PREC_CUTOFF */
};

static const char *redir[] =
{
	"(%s) redirect %s to network %s",	/* ICMP_REDIR_NET */
	"(%s) redirect %s to host %s",	/* ICMP_REDIR_HOST */
	"(%s) redirect-tos %s to network %s",	/* ICMP_REDIR_NETTOS */
	"(%s) redirect-tos %s to host %s",	/* ICMP_REDIR_HOSTTOS */
};

int main(void)
{
	int sock;
	struct in_addr in;
	struct packet pkt;

#ifdef DETECT_SMURF
	signal(SIGALRM, sigalrm);
#endif

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("icmplog", LOGGING_PRIORITY, LOGGING_FACILITY);
	signal(SIGHUP, SIG_IGN);

	fork_to_background();
	initialize_lockfile("icmplog");
	drop_privileges();

	for (;;) {
		read(sock, (struct packet *)&pkt, sizeof(pkt));

		if (pkt.ip.ip_hl != 5) {
			syslog(LEVEL, "suspicious ip options from %s", inet_ntoa(src));
			continue;
		}
		switch (pkt.icmp.icmp_type) {
			case ICMP_TIME_EXCEEDED:
				break;			/* ignored, because traceroute(1) makes logs messy */
			case ICMP_ECHOREPLY:
#ifdef DETECT_SMURF
				if (smurf(pkt))
					syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
#endif
			case ICMP_ECHO:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
			case ICMP_DEST_UNREACH:
				syslog(LEVEL, unreach[(int)pkt.icmp.icmp_code], inet_ntoa(src), protolookup(proto), UNFIX(pkt.tcp.th_dport), UNFIX(pkt.tcp.th_sport));
				break;
			case ICMP_REDIRECT:
				if (gw.s_addr != src.s_addr)
					syslog(LEVEL, "warning: gateway address (%s) != source address (%s)", my_ntoa(gw), my_ntoa(src));
				syslog(LEVEL, redir[(int)pkt.icmp.icmp_code], my_ntoa(gw), my_ntoa(orig), my_ntoa(dest));
				break;
			case ICMP_TIMESTAMPREPLY:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), print_time(pkt.icmp.icmp_ttime));
				break;
			case ICMP_ADDRESSREPLY:
				in.s_addr = htonl(pkt.icmp.icmp_mask);
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], my_ntoa(src), my_ntoa(in));
				break;
			case ICMP_PARAMETERPROB:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), pkt.icmp.icmp_pptr);
				break;
			default:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src));
				break;
		}
	}
}

static char *protolookup(int num)
{
	struct protoent *pe;

	if ((pe = getprotobynumber(num)) != NULL)
		return (pe->p_name);
	else {
		static char buff[16];

		sprintf(buff, "proto %d", num);
		return (buff);
	}
}

/* inet_ntoa() uses a statically allocated variable. bah. */
static char *my_ntoa(struct in_addr in)
{
	char host[32];
	unsigned char *ip;

	ip = (unsigned char *)&in;
	sprintf(host, "%d.%d.%d.%d", (ip[0] & 255), (ip[1] & 255), (ip[2] & 255), (ip[3] & 255));

	return (strdup(host));
}

static char *print_time(u_int time)
{
	struct tm *tm;
	char ret[32];
	u_long actual_time;

	actual_time = ntohl(time) / 1000;
	tm = localtime((time_t *) & actual_time);
	sprintf(ret, "%.2d:%.02d:%.02d", tm->tm_hour, tm->tm_min, tm->tm_sec);

	return (strdup(ret));
}
